#include <iostream>

#include <map>
#include <vector>
#include <sstream>

#include <stdio.h>
#include <string>
#include "ch.hh"
#include "c4spd/c4spd.hh"

static std::string filename;
static unsigned    trackno;
static double FPS = 60;

class NumStat
{
    unsigned n;
    double min,max,sum;
    double sqr_sum;
public:
    NumStat(): n(0),min(0),max(0),sum(0) { }
    void Cumulate(double val)
    {
        sum+=val; sqr_sum += val*val;
        if(!n||val<min)min=val;
        if(!n||val>max)max=val;
        ++n;
    }
    double GetAverage() const { return sum/n; }
    double GetMin() const { return min; }
    double GetMax() const { return max; }
    unsigned GetCount() const { return n; }
    double GetVariance() const { return (sqr_sum - sum*GetAverage()) / (n-1); }
};

typedef std::map<unsigned/*chn*/,int/*tone*/> chstatus_t;
typedef std::map<unsigned/*time*/, chstatus_t> melody_t;

static const std::string NameTone(int tone)
{
    static const char* const tonenames[12] =
    {
        "C-","C#","D-","D#","E-","F-","F#","G-","G#","A-","A#","B-"
    };
    char result[16];
    sprintf(result, "%s%d", tonenames[(tone)%12], (tone)/12-6);
    return result;
}

static const std::string sqlfix(const std::string &q)
{
    const char *s = q.c_str();
    std::string diu = "'";
    while(*s)
    {        
        if(*s=='\'' || *s=='\\')diu += '\\';
        diu += *s;
        ++s;
    }
    return diu + "'";
}

static void SaveMelody(const melody_t& track, unsigned loop_begin, unsigned track_size)
{
    const double fps = FPS;
    
    std::map<int, std::vector<int> > channels_rel;
    std::map<int, std::vector<int> > channels_abs;
    std::map<int, std::vector<unsigned> > timings;
    
    {std::map<int, int> prevtones;
    for(melody_t::const_iterator i = track.begin(); i != track.end(); ++i)
    {
        unsigned timepos = i->first;
        for(chstatus_t::const_iterator j = i->second.begin(); j != i->second.end(); ++j)
        {
            int chno = j->first;

            //fprintf(stderr, "ch %u t %u: %d\n", chno, timepos, (int)j->second);

            int tonediff = j->second - prevtones[chno];
            prevtones[chno] = j->second;
            channels_rel[chno].push_back(tonediff);
            channels_abs[chno].push_back(j->second);
            timings[chno].push_back(timepos);
        }
    }}
    
    std::stringstream outbuffer;
    
    outbuffer<< "insert into song set filename=" << sqlfix(filename)
             << ",trackno=" << trackno
             << ",loopbegin=" << (loop_begin / fps)
             << ",length=" << (track_size / fps)
             << ";\n";

    outbuffer<< "set @songid = last_insert_id();\n";

    std::map<int, std::vector<int> >::const_iterator ai = channels_abs.begin();
    for(std::map<int, std::vector<int> >::const_iterator
        i = channels_rel.begin(); i != channels_rel.end(); ++i, ++ai)
    {
        int chno = i->first;
        std::string data_rel, data_abs;

        std::vector<unsigned> rel_curspos;
        
        { const std::vector<int>& trk_rel = i->second;
        rel_curspos.reserve(trk_rel.size());
        for(unsigned a=0; a<trk_rel.size(); ++a)
        {
            rel_curspos.push_back(data_rel.size());
            data_rel += (char)(128 + trk_rel[a]);
        }}
        
        std::vector<unsigned> abs_curspos;
        
        { const std::vector<int>& trk_abs = ai->second;
        abs_curspos.reserve(trk_abs.size()*2);
        int prev_oct = 0;
        for(unsigned a=0; a<trk_abs.size(); ++a)
        {
            int note = trk_abs[a], oct = note/12;
            while(oct < prev_oct) { data_abs += '<'; --prev_oct; }
            while(oct > prev_oct) { data_abs += '>'; ++prev_oct; }
            abs_curspos.push_back(data_abs.size());
            data_abs += (char)(65 + note%12);
        }}
        
        outbuffer<< "insert into song_channel values(@songid, "
                 << chno
                 << "," << sqlfix(data_rel)
                 << "," << sqlfix(data_abs)
                 << ");\n";
        
        for(unsigned b=0; b<data_rel.size(); ++b)
        {
            outbuffer<< "insert into song_index values(@songid, " << chno
                     << "," << b
                     << "," << (timings[chno][b] / fps)
                     << "," << sqlfix(data_rel.substr(rel_curspos[b], 32))
                     << "," << sqlfix(data_abs.substr(abs_curspos[b], 32))
                     << ");\n";
        }
    }
    std::cout << outbuffer.str() << std::flush;
}

static void SaveMelody(const std::vector<situation>& track, unsigned loop_begin)
{
    std::map<unsigned, NumStat> melodic_deviation;
    unsigned max_channel=0;
    
    for(unsigned a=0; a<track.size(); ++a)
        for(situation::const_iterator
            i = track[a].begin();
            i != track[a].end();
            ++i)
        {
            if(i->first > max_channel) max_channel = i->first;
            double tone = fiddle_ftom(i->second.hz);
            double difference = fmod(tone, 1.0);
            if(difference > 0.5) difference -= 1.0;
            melodic_deviation[i->first].Cumulate(difference);
        }
    for(std::map<unsigned, NumStat>::const_iterator
        j = melodic_deviation.begin();
        j != melodic_deviation.end();
        ++j)
    {
        fprintf(stderr,
            "Trk %u: deviation %lf semitones (%lf..%lf), variance %lf from standard tuning (%u samples)\n",
            j->first,
            j->second.GetAverage(),  j->second.GetMin(), j->second.GetMax(),
            j->second.GetVariance(),
            j->second.GetCount()
        );
    }
    
    std::vector<double> prev_tone(max_channel+1);
    
    melody_t melody_simplified;
    for(unsigned a=0; a<track.size(); ++a)
    {
        for(situation::const_iterator
            i = track[a].begin();
            i != track[a].end();
            ++i)
        {
            double tone = fiddle_ftom(i->second.hz)
                          - melodic_deviation[i->first].GetAverage();
                    ;

            const double abs_tonediff = fabs(prev_tone[i->first] - tone);
            prev_tone[i->first] = tone;
            
            double prev_volume = 0.0;
            if(a > 0) { situation::const_iterator j = track[a-1].find(i->first);
                        if(j != track[a-1].end()) prev_volume = j->second.amplitude; }
            
            double volume = i->second.amplitude;
            
            if(prev_volume < silence_level
            || volume >= prev_volume + 2/16.0
            || (prev_volume >= silence_level && volume >= prev_volume*3)
            || abs_tonediff >= 0.92856
            || a == loop_begin)
            {
                // It's a tick point.
                melody_simplified[a][i->first] = (int)(tone+0.5);
            }
        }
    }
    
    SaveMelody(melody_simplified, loop_begin, track.size());
}

int main(int argc, char**argv)
{
    filename = argv[1];
    trackno  = std::atoi(argv[2]);

    if(scanf("%lf", &FPS) != 1) return -1;

    unsigned track_size, loop_begin;
    if(scanf("%u %u", &track_size, &loop_begin) != 2) return -1;
    
    std::vector<situation> track(track_size);
    for(unsigned a=0; a<track.size(); ++a)
    {
        unsigned num_chs;
        if(scanf("%u", &num_chs) != 1) break;
        situation line;
        for(unsigned n=0; n<num_chs; ++n)
        {
            unsigned chno;
            ch ch;
            if(scanf("%u %lf %lf", &chno, &ch.amplitude, &ch.hz) != 3) break;
            line[chno] = ch;
        }
        track[a] = line;
    }
    SaveMelody(track, loop_begin);
}
